//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <algorithm>
#include <sstream>
#include <stdexcept>

#include "lfc1/filesystem/cdirectory.hpp"
#include "lfc1/system/cerrorcategory.hpp"

using namespace std;
using namespace lfc1::system;
using namespace boost::filesystem;

namespace lfc1 { namespace filesystem {

///	@brief This function creates a CDirectory object.
///
///	@param[in]	arPath	The name of the directory.
///
CDirectory::CDirectory(const boost::filesystem::path& arPath)
:	mvPath(arPath),
	mvEntries()
{
	// Load and sort the contents of the directory.
	mvRefresh();
}

///	@brief This function copy constructs a CDirectory object.
///
///	@param[in]	arRHS	The CDirectory object to be copied.
///
CDirectory::CDirectory(const CDirectory& arRHS)
:	mvPath(arRHS.mvPath),
	mvEntries(arRHS.mvEntries)
{
}

///	@brief This function move constructs a CDirectory object.
///
///	@param[in]	arRHS	The CDirectory object to be moved.
///
CDirectory::CDirectory(CDirectory&& arRHS)
:	mvPath(),
	mvEntries()
{
	*this = std::move(arRHS);
}

///	@brief This function destroys a CDirectory object.
///
CDirectory::~CDirectory() noexcept
{
}

///	@brief This function assigns a CDirectory object to another CDirectory object.
///
///	@param[in]	arRHS	The CDirectory object to assign to another CDirectory object.
///
///	@return	The CDirectory object assigned to.
///
CDirectory& CDirectory::operator=(const CDirectory& arRHS)
{
	CDirectory vCopy(arRHS);
	swap(vCopy);
	return *this;
}

///	@brief This function moves a CDirectory object to another CDirectory object.
///
///	@param[in]	arRHS	The CDirectory object to move to another CDirectory object.
///
///	@return	The CDirectory object assigned to.
///
CDirectory& CDirectory::operator=(CDirectory&& arRHS) noexcept
{
	swap(arRHS);
	return *this;
}

///	@brief This function obtains the name of the directory.
///
///	@return The name of the directory.
///
boost::filesystem::path CDirectory::mvGetPath() const
{
	return mvPath;
}

///	@brief This function obtains the number of directory entries.
///
///	@return The number of directory entries.
///
unsigned CDirectory::mvGetNumEntries() const noexcept
{
	return static_cast<unsigned>(mvEntries.size());
}

///	@brief This function obtains the beginning iterator to the list of directory entries.
///
///	@return The beginning iterator to the list of directory entries.
///
///	@note This function does not follow this library's member function naming convention in order to support
///		  range-base for loops.
///
CDirectory::TIterator CDirectory::begin() const noexcept
{
	return mvEntries.begin();
}

///	@brief This function obtains the ending iterator to the list of directory entries.
///
///	@return The ending iterator to the list of directory entries.
///
///	@note This function does not follow this library's member function naming convention in order to support
///		  range-base for loops.
///
CDirectory::TIterator CDirectory::end() const noexcept
{
	return mvEntries.end();
}

///	@brief This function swaps a CDirectory object with another CDirectory object.
///
///	@param[in,out]	arRHS	The CDirectory object to swap with.
///
void CDirectory::swap(CDirectory& arRHS) noexcept
{
	using std::swap;
	swap(mvPath, arRHS.mvPath);
	swap(mvEntries, arRHS.mvEntries);
}

///	@brief This function loads and sorts the contents of a directory.
///
///	@exception	std::system_error	Indicates a directory open or read failure.
///
void CDirectory::mvRefresh()
{
	// Create a temporary list of directory entries.
	TEntries vEntries;

	// Open the directory.
	directory_iterator pDirIt;
	try
	{
		pDirIt = directory_iterator(mvPath);
	}
	catch (filesystem_error&)
	{
		ostringstream vStream;
		vStream
			<< "Unable to open directory '" << mvPath.string() << "' at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw system_error(E_EC_DIR_OPEN_ERROR, grCategory(), vStream.str());
	}

	// Read the directory one entry at a time.
	while (pDirIt != directory_iterator())
	{
		// Add the directory entry to our list.
		vEntries.push_back(*pDirIt);

		// Move on to the next directory entry.
		try
		{
			++pDirIt;
		}
		catch (filesystem_error&)
		{
			ostringstream vStream;
			vStream
				<< "Unable to read directory '" << mvPath.string() << "' at ("
				<< __FILE__ << ", " << __LINE__ << ")";
			throw system_error(E_EC_DIR_READ_ERROR, grCategory(), vStream.str());
		}
	}

	// Sort the completed list.
	if (!vEntries.empty())
	{
		sort(vEntries.begin(), vEntries.end());
	}

	// Swap our temporary list with the final list.
	std::swap(mvEntries, vEntries);
}

///	@ingroup LFC1_FILESYSTEM
///	@headerfile lfc1/filesystem/cdirectory.hpp
///	@brief This function overloads std::swap for CDirectory objects.
///
///	@param[in,out]	arLHS	The first object to swap.
///	@param[in,out]	arRHS	The second object to swap.
///
void swap(CDirectory& arLHS, CDirectory& arRHS) noexcept
{
	arLHS.swap(arRHS);
}

} /* namespace filesystem */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
