//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#include <algorithm>
#include <sstream>
#include <stdexcept>

#include "lfc1/filesystem/crecdirectory.hpp"
#include "lfc1/system/cerrorcategory.hpp"

using namespace std;
using namespace lfc1::system;
using namespace boost::filesystem;

namespace lfc1 { namespace filesystem {

///	@brief This function creates a default CRecDirectory object.
///
///	@param[in]	arPath	The name of the directory.
///
CRecDirectory::CRecDirectory(const boost::filesystem::path& arPath)
:	mvLevelCount(0),
	mvPath(arPath),
	mvEntries()
{
	// Load and sort the contents of the directory.
	mvRefresh();
}

///	@brief This function copy constructs a CRecDirectory object.
///
///	@param[in]	arRHS	The CRecDirectory object to be copied.
///
CRecDirectory::CRecDirectory(const CRecDirectory& arRHS)
:	mvLevelCount(arRHS.mvLevelCount),
	mvPath(arRHS.mvPath),
	mvEntries(arRHS.mvEntries)
{
}

///	@brief This function move constructs a CRecDirectory object.
///
///	@param[in]	arRHS	The CRecDirectory object to be moved.
///
CRecDirectory::CRecDirectory(CRecDirectory&& arRHS)
:	mvLevelCount(0),
	mvPath(),
	mvEntries()
{
	*this = std::move(arRHS);
}

///	@brief This function destroys a CRecDirectory object.
///
CRecDirectory::~CRecDirectory() noexcept
{
}

///	@brief This function assigns a CRecDirectory object to another CRecDirectory object.
///
///	@param[in]	arRHS	The CRecDirectory object to assign to another CRecDirectory object.
///
///	@return	The CRecDirectory object assigned to.
///
CRecDirectory& CRecDirectory::operator=(const CRecDirectory& arRHS)
{
	CRecDirectory vCopy(arRHS);
	swap(vCopy);
	return *this;
}

///	@brief This function moves a CRecDirectory object to another CRecDirectory object.
///
///	@param[in]	arRHS	The CRecDirectory object to move to another CRecDirectory object.
///
///	@return	The CRecDirectory object assigned to.
///
CRecDirectory& CRecDirectory::operator=(CRecDirectory&& arRHS) noexcept
{
	swap(arRHS);
	return *this;
}

///	@brief This function obtains the number of directory levels.
///
///	@return The number of directory levels.
///
int CRecDirectory::mvGetLevelCount() const noexcept
{
	return mvLevelCount;
}

///	@brief This function obtains the name of the directory.
///
///	@return The name of the directory.
///
boost::filesystem::path CRecDirectory::mvGetPath() const
{
	return mvPath;
}

///	@brief This function obtains the number of directory entries.
///
///	@return The number of directory entries.
///
unsigned CRecDirectory::mvGetNumEntries() const noexcept
{
	return static_cast<unsigned>(mvEntries.size());
}

///	@brief This function obtains the beginning iterator to the list of directory entries.
///
///	@return The beginning iterator to the list of directory entries.
///
///	@note This function does not follow this library's member function naming convention in order to support
///		  range-base for loops.
///
CRecDirectory::TIterator CRecDirectory::begin() const noexcept
{
	return mvEntries.begin();
}

///	@brief This function obtains the ending iterator to the list of directory entries.
///
///	@return The ending iterator to the list of directory entries.
///
///	@note This function does not follow this library's member function naming convention in order to support
///		  range-base for loops.
///
CRecDirectory::TIterator CRecDirectory::end() const noexcept
{
	return mvEntries.end();
}

///	@brief This function swaps a CRecDirectory object with another CRecDirectory object.
///
///	@param[in,out]	arRHS	The CRecDirectory object to swap with.
///
void CRecDirectory::swap(CRecDirectory& arRHS) noexcept
{
	using std::swap;
	swap(mvLevelCount, arRHS.mvLevelCount);
	swap(mvPath, arRHS.mvPath);
	swap(mvEntries, arRHS.mvEntries);
}

///	@brief This function loads and sorts the contents of a directory.
///
///	@exception	std::system_error	Indicates a directory open or read failure.
///
void CRecDirectory::mvRefresh()
{
	// Create a temporary list of directory entries.
	int vLevelCount = 0;
	TEntries vEntries;

	// Open the directory.
	recursive_directory_iterator pDirIt;
	try
	{
		pDirIt = recursive_directory_iterator(mvPath);
	}
	catch (filesystem_error&)
	{
		ostringstream vStream;
		vStream
			<< "Unable to open directory '" << mvPath.string() << "' at ("
			<< __FILE__ << ", " << __LINE__ << ")";
		throw system_error(E_EC_DIR_OPEN_ERROR, grCategory(), vStream.str());
	}

	// Read the directory one entry at a time.
	while (pDirIt != recursive_directory_iterator())
	{
		// Add the directory entry to our list.
		TEntry vEntry(*pDirIt, pDirIt.level());
		vEntries.push_back(vEntry);
		if (pDirIt.level() > vLevelCount)
		{
			vLevelCount = pDirIt.level();
		}

		// Move on to the next directory entry.
		try
		{
			++pDirIt;
		}
		catch (filesystem_error&)
		{
			ostringstream vStream;
			vStream
				<< "Unable to read directory '" << mvPath.string() << "' at ("
				<< __FILE__ << ", " << __LINE__ << ")";
			throw system_error(E_EC_DIR_READ_ERROR, grCategory(), vStream.str());
		}
	}

	// Sort the completed list.
	if (!vEntries.empty())
	{
		sort(vEntries.begin(), vEntries.end());
	}

	// Swap our temporary list with the final list.
	std::swap(mvLevelCount, ++vLevelCount);
	std::swap(mvEntries, vEntries);
}

///	@ingroup LFC1_FILESYSTEM
///	@headerfile lfc1/filesystem/crecdirectory.hpp
///	@brief This function overloads std::swap for CRecDirectory objects.
///
///	@param[in,out]	arLHS	The first object to swap.
///	@param[in,out]	arRHS	The second object to swap.
///
void swap(CRecDirectory& arLHS, CRecDirectory& arRHS) noexcept
{
	arLHS.swap(arRHS);
}

} /* namespace filesystem */ } /* namespace lfc1 */

//============================================================================================================
//	End of Code
//============================================================================================================
