//============================================================================================================
//	$Id$
//
//	$Author$
//
//	$DateTime$
//============================================================================================================

#ifndef _LFC1_NUMERIC_CINTEGER_HPP_
#define _LFC1_NUMERIC_CINTEGER_HPP_

#include <algorithm>
#include <limits>
#include <type_traits>

#include "lfc1/numeric/numericexp.hpp"

namespace lfc1 { namespace numeric {

///	@ingroup LFC1_NUMERIC
///	@headerfile lfc1/numeric/cinteger.hpp
///	@brief This class template represents little endian integers on big endian operating systems or big endian
///		   integers on little endian operating systems.
///
///	This class supports the following operators:
///	@code
///	binary arithmetic operators: +=, -=, *=, /=
///	unary arithmetic operators:  ++, --
///	binary bitwise operators:    &=, |=, ^=
///	shift operators:             <<=, >>=
///	@endcode
///
///	@tparam	IT	The integer type.
///
template <typename IT>
class CInteger
{
private:
	// Template argument must be a 2, 4 or 8-byte integer type.
	static_assert(std::is_integral<IT>::value, "Must be an integer");
	static_assert(sizeof(IT) > sizeof(char), "Must be larger than a character");

public:
	// Constructors/destructor
	CInteger() noexcept;
	CInteger(IT avInteger) noexcept;
	CInteger(const CInteger& arRHS) noexcept;
	CInteger(CInteger&& arRHS) noexcept;
	~CInteger() noexcept;

	// Assignment
	CInteger& operator=(IT avInteger) noexcept;
	CInteger& operator=(const CInteger& arRHS) noexcept;
	CInteger& operator=(CInteger&& arRHS) noexcept;

	// Conversions
	operator IT () const noexcept;

	// Instance special operations
	void swap(CInteger& arRHS) noexcept;
	CInteger& operator++() noexcept;
	CInteger operator++(int) noexcept;
	CInteger& operator--() noexcept;
	CInteger operator--(int) noexcept;
	CInteger& operator+=(const CInteger& arRHS) noexcept;
	CInteger& operator-=(const CInteger& arRHS) noexcept;
	CInteger& operator*=(const CInteger& arRHS) noexcept;
	CInteger& operator/=(const CInteger& arRHS) noexcept;
	CInteger& operator&=(const CInteger& arRHS) noexcept;
	CInteger& operator|=(const CInteger& arRHS) noexcept;
	CInteger& operator^=(const CInteger& arRHS) noexcept;
	CInteger& operator<<=(const CInteger& arRHS) noexcept;
	CInteger& operator>>=(const CInteger& arRHS) noexcept;

private:
	// Instance data
	IT mvData;
};

///	@brief This template function creates a default CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
template <typename IT>
CInteger<IT>::CInteger() noexcept
{
}

///	@brief This template function creates a CInteger<IT> object and assigns the given value to it.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	avInteger	The value to assign to the CInteger<IT> object.
///
template <typename IT>
CInteger<IT>::CInteger(IT avInteger) noexcept
{
	mvData = avInteger;
	auto pData = reinterpret_cast<unsigned char*>(&mvData);
	switch (sizeof(mvData))
	{
	case 2:
		std::swap(*(pData + 0), *(pData + 1));
		break;
	case 4:
		std::swap(*(pData + 0), *(pData + 3));
		std::swap(*(pData + 1), *(pData + 2));
		break;
	case 8:
		std::swap(*(pData + 0), *(pData + 7));
		std::swap(*(pData + 1), *(pData + 6));
		std::swap(*(pData + 2), *(pData + 5));
		std::swap(*(pData + 3), *(pData + 4));
		break;
	default:
		break;
	}
}

///	@brief This template function copy constructs a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The CInteger<IT> object to be copied.
///
template <typename IT>
CInteger<IT>::CInteger(const CInteger& arRHS) noexcept
:	mvData(arRHS.mvData)
{
}

///	@brief This template function move constructs a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The CInteger<IT> object to be moved.
///
template <typename IT>
CInteger<IT>::CInteger(CInteger&& arRHS) noexcept
:	mvData(arRHS.mvData)
{
}

///	@brief This template function destroys a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
template <typename IT>
CInteger<IT>::~CInteger() noexcept
{
}

///	@brief This template function assigns the given value to a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	avInteger	The value to assign to the CInteger<IT> object.
///
///	@return	The CInteger<IT> object assigned to.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator=(IT avInteger) noexcept
{
	CInteger vCopy(avInteger);
	swap(vCopy);
	return *this;
}

///	@brief This template function assigns a CInteger<IT> object to another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The CInteger<IT> object to assign to another CInteger<IT> object.
///
///	@return	The CInteger<IT> object assigned to.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator=(const CInteger& arRHS) noexcept
{
	mvData = arRHS.mvData;
	return *this;
}

///	@brief This template function moves a CInteger<IT> object to another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The CInteger<IT> object to move to another CInteger<IT> object.
///
///	@return	The CInteger<IT> object assigned to.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator=(CInteger&& arRHS) noexcept
{
	mvData = arRHS.mvData;
	return *this;
}

///	@brief This template function converts a CInteger<IT> object to its corresponding integer type.
///
///	@tparam	IT	The integer type.
///
///	@return	The value of the integer type.
///
template <typename IT>
CInteger<IT>::operator IT () const noexcept
{
	IT vInteger = mvData;
	auto pInteger = reinterpret_cast<unsigned char*>(&vInteger);
	switch (sizeof(mvData))
	{
	case 2:
		std::swap(*(pInteger + 0), *(pInteger + 1));
		break;
	case 4:
		std::swap(*(pInteger + 0), *(pInteger + 3));
		std::swap(*(pInteger + 1), *(pInteger + 2));
		break;
	case 8:
		std::swap(*(pInteger + 0), *(pInteger + 7));
		std::swap(*(pInteger + 1), *(pInteger + 6));
		std::swap(*(pInteger + 2), *(pInteger + 5));
		std::swap(*(pInteger + 3), *(pInteger + 4));
		break;
	default:
		break;
	}
	return vInteger;
}

///	@brief This template function swaps a CInteger<IT> object with another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in,out]	arRHS	The CInteger<IT> object to swap with.
///
template <typename IT>
void CInteger<IT>::swap(CInteger& arRHS) noexcept
{
	using std::swap;
	swap(mvData, arRHS.mvData);
}

///	@brief This template function pre-increments a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator++() noexcept
{
	IT vInteger = *this;
	*this = ++vInteger;
	return *this;
}

///	@brief This template function post-increments a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@return	The value of the CInteger<IT> object before the increment.
///
template <typename IT>
CInteger<IT> CInteger<IT>::operator++(int) noexcept
{
	IT vInteger = *this;
	++(*this);
	return vInteger;
}

///	@brief This template function pre-decrements a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator--() noexcept
{
	IT vInteger = *this;
	*this = --vInteger;
	return *this;
}

///	@brief This template function post-decrements a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@return	The value of the CInteger<IT> object before the decrement.
///
template <typename IT>
CInteger<IT> CInteger<IT>::operator--(int) noexcept
{
	IT vInteger = *this;
	--(*this);
	return vInteger;
}

///	@brief This template function adds another CInteger<IT> object to a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator+=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this + arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function subtracts another CInteger<IT> object from a CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator-=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this - arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function multiplies a CInteger<IT> object by another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator*=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this * arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function divides a CInteger<IT> object by another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator/=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this / arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function logical ANDs a CInteger<IT> object with another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator&=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this & arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function logical ORs a CInteger<IT> object with another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator|=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this | arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function logical XORs a CInteger<IT> object with another CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator^=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this ^ arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function left shifts a CInteger<IT> object by the number of bits given by another
///		   CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator<<=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this << arRHS;
	*this = vInteger;
	return *this;
}

///	@brief This template function right shifts a CInteger<IT> object by the number of bits given by another
///		   CInteger<IT> object.
///
///	@tparam	IT	The integer type.
///
///	@param[in]	arRHS	The other CInteger<IT> object.
///
///	@return	The CInteger<IT> object.
///
template <typename IT>
CInteger<IT>& CInteger<IT>::operator>>=(const CInteger& arRHS) noexcept
{
	IT vInteger = *this >> arRHS;
	*this = vInteger;
	return *this;
}

///	@ingroup LFC1_NUMERIC
///	@headerfile lfc1/numeric/cinteger.hpp
///	@brief This template function overloads std::swap for CInteger<IT> objects.
///
///	@tparam	IT	The integer type.
///
///	@param[in,out]	arLHS	The first object to swap.
///	@param[in,out]	arRHS	The second object to swap.
///
template <typename IT>
void swap(CInteger<IT>& arLHS, CInteger<IT>& arRHS) noexcept
{
	arLHS.swap(arRHS);
}

} /* namespace numeric */ } /* namespace lfc1 */

namespace std {

///	@ingroup LFC1_NUMERIC
///	@headerfile lfc1/numeric/cinteger.hpp
///	@brief This template class is a specialization for std::numeric_limits for CInteger<IT>.
///
///	@tparam	IT	The integer type.
///
template <typename IT>
class numeric_limits<lfc1::numeric::CInteger<IT>>
{
public:
	static const bool is_specialized = true;	///< Have information for this type.

	static lfc1::numeric::CInteger<IT> min() noexcept	///< Lowest possible value.
	{
		return numeric_limits<IT>::min();
	}
	static lfc1::numeric::CInteger<IT> max() noexcept	///< Highest possible value.
	{
		return numeric_limits<IT>::max();
	}
	static const int digits = numeric_limits<IT>::digits;	///< Number of significant bits.
	static const int digits10 = numeric_limits<IT>::digits10;	///< Number of significant digits.
	static const bool is_signed = numeric_limits<IT>::is_signed;	///< Has sign or not.
	static const bool is_integer = true;	///< Is a whole number.
	static const bool is_exact = true;	///< Is an exact number.
	static const int radix = 2;	///< Uses binary representation.
	static lfc1::numeric::CInteger<IT> epsilon() noexcept { return 0; }	///< Not applicable.
	static lfc1::numeric::CInteger<IT> round_error() noexcept { return 0; }	///< Not applicable.
	static const int min_exponent = 0;	///< Not applicable.
	static const int min_exponent10 = 0;	///< Not applicable.
	static const int max_exponent = 0;	///< Not applicable.
	static const int max_exponent10 = 0;	///< Not applicable.
	static const bool has_infinity = false;	///< Not applicable.
	static const bool has_quiet_NaN = false;	///< Not applicable.
	static const bool has_signaling_NaN = false;	///< Not applicable.
	static const float_denorm_style has_denorm = denorm_absent;	///< Not applicable.
	static const bool has_denorm_loss = false;	///< Not applicable.
	static lfc1::numeric::CInteger<IT> infinity() noexcept { return 0; }	///< Not applicable.
	static lfc1::numeric::CInteger<IT> quiet_NaN() noexcept { return 0; }	///< Not applicable.
	static lfc1::numeric::CInteger<IT> signaling_NaN() noexcept { return 0; }	///< Not applicable.
	static lfc1::numeric::CInteger<IT> denorm_min() noexcept { return 0; }	///< Not applicable.
	static const bool is_iec559 = false;	///< Not applicable.
	static const bool is_bounded = true;	///< Has a range of valid values.
	static const bool is_modulo = true;	///< Values wrap around on underflow/overflow.
	static const bool traps = false;	///< Not applicable.
	static const bool tinyness_before = false;	///< Not applicable.
	static const float_round_style round_style = round_toward_zero;	///< Not applicable.
};

} /* namespace std */

#endif // _LFC1_NUMERIC_CINTEGER_HPP_

//============================================================================================================
//	End of Code
//============================================================================================================
